from operator import attrgetter
import unittest

from pyzusi3 import messages


class TestOperators(unittest.TestCase):
    def test_parameter_operators(self):
        param1 = messages.ParameterId(id1=2, id2=266, id3=None, id4=None, id5=None, id6=None)
        param2 = messages.ParameterId(id1=2, id2=266, id3=1, id4=None, id5=None, id6=None)
        self.assertTrue(param1 < param2)
        self.assertTrue(param2 > param1)
        param1 = messages.ParameterId(id1=2, id2=266, id3=1, id4=1, id5=None, id6=None),
        param2 = messages.ParameterId(id1=2, id2=266, id3=2, id4=None, id5=None, id6=None),
        self.assertTrue(param1 < param2)
        self.assertTrue(param2 > param1)
        param1 = messages.ParameterId(id1=2, id2=266, id3=1, id4=None, id5=None, id6=None)
        param2 = messages.ParameterId(id1=2, id2=266, id3=1, id4=1, id5=None, id6=None)
        self.assertTrue(param1 < param2)
        self.assertTrue(param2 > param1)
        param1 = messages.ParameterId(id1=2, id2=266, id3=2, id4=2, id5=4, id6=1)
        param2 = messages.ParameterId(id1=2, id2=266, id3=12, id4=2, id5=None, id6=None)
        self.assertTrue(param1 < param2)
        self.assertTrue(param2 > param1)

    def test_parameter_lt_sort(self):
        unsorted = (
            messages.ParameterId(id1=2, id2=266, id3=1, id4=1, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=2, id4=None, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=None, id4=None, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=1, id4=None, id5=None, id6=None),
        )
        real_sorted = (
            messages.ParameterId(id1=2, id2=266, id3=None, id4=None, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=1, id4=None, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=1, id4=1, id5=None, id6=None),
            messages.ParameterId(id1=2, id2=266, id3=2, id4=None, id5=None, id6=None),
        )
        sorting_test = tuple(sorted(unsorted))
        self.assertEqual(sorting_test, real_sorted)


class TestInstantiation(unittest.TestCase):
    def test_all_messages(self):
        """Ensure all messages can be instantiated without parameters"""

        messages.ParameterId()
        with self.assertRaises(TypeError):
            # needs 3 parameters at least
            messages.LowlevelParameter()
        messages.HELLO()
        messages.ACK_HELLO()
        messages.NEEDED_DATA()
        messages.ACK_NEEDED_DATA()
        messages.DATA_FTD()
        messages.STATUS_NOTBREMSSYSTEM()
        messages.STATUS_SIFA()
        messages.STATUS_ZUGBEEINFLUSSUNG_GRUND()
        messages.STATUS_INDUSI_EINSTELLUNGEN()
        messages.STATUS_INDUSI_BETRIEBSDATEN()
        messages.STATUS_ETCS_EINSTELLUNGEN()
        messages.STATUS_ETCS_EINSTELLUNGEN_STMSYSTEM()
        messages.STATUS_ETCS_BETRIEBSDATEN()
        messages.STATUS_ETCS_BETRIEBSDATEN_VORSCHAUPUNKT()
        messages.STATUS_ZUB_EINSTELLUNGEN()
        messages.STATUS_ZUB_BETRIEBSDATEN()
        messages.STATUS_ZBS_EINSTELLUNGEN()
        messages.STATUS_ZBS_EINSTELLUNGEN_ALTSYSTEM()
        messages.STATUS_ZBS_BETRIEBSDATEN()
        messages.STATUS_FAHRSPERRE_EINSTELLUNGEN()
        messages.STATUS_FAHRSPERRE_BETRIEBSDATEN()
        messages.STATUS_TUEREN()
        messages.STATUS_FAHRZEUG()
        messages.STATUS_ZUGVERBAND()
        messages.STATUS_WEICHEN()
        messages.STATUS_WEICHEN_WEICHE()
        messages.STATUS_LM_ZUSIDISPLAY()
        messages.STATUS_LM_ZUSIDISPLAY_DISPLAY()
        messages.STATUS_ZUGFAHRDATEN()
        messages.STATUS_ZUGFAHRDATEN_FAHRZEUG()
        messages.DATA_OPERATION()
        messages.DATA_OPERATION_BETAETIGUNG()
        messages.DATA_OPERATION_KOMBISCHALTER()
        messages.DATA_OPERATION_SCHALTERFUNKTION()
        messages.DATA_OPERATION_MAUSKLICK()
        messages.DATA_PROG()
        messages.INPUT()
        messages.CONTROL()



import asyncio
import logging
import os.path
import sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from pyzusi3 import messages
from pyzusi3.client import ZusiClient, suppress_none_values

ZUSI_IP = "127.0.0.1"
ZUSI_PORT = "1436"

WAIT_MAX = 3000 # keep running for 5 minutes

logging.basicConfig(level=logging.WARNING)
log = logging.getLogger("ZusiDemo")
log.setLevel(logging.INFO)
logging.getLogger("pyzusi3.client").setLevel(logging.INFO)
logging.getLogger("pyzusi3.messagecoders.MessageDecoder").setLevel(logging.ERROR)


def handle_exception(loop, context):
    """Dummy handler to log errors in demo to console"""
    msg = context.get("exception", context["message"])
    log.error(f"Caught exception: {msg}")


async def zusi_user_interact(client: ZusiClient):
    """Simulation for some random user input to be sent to Zusi"""

    zuord = messages.INPUT_TASTATURZUORDNUNG
    komm = messages.INPUT_TASTATURKOMMANDO
    aktion = messages.INPUT_TASTATURAKTION

    simulated_interactions = {
        100: messages.INPUT(zuord.FAHRSCHALTER, komm.Unbestimmt, aktion.Absolut, tastatur_schalterposition=5),
        109: messages.INPUT(zuord.SIFA, komm.Unbestimmt, aktion.Down, tastatur_schalterposition=0),
        110: messages.INPUT(zuord.SIFA, komm.Unbestimmt, aktion.Up, tastatur_schalterposition=0),
        150: messages.INPUT(zuord.FAHRSCHALTER, komm.Unbestimmt, aktion.Absolut, tastatur_schalterposition=0),
        160: messages.INPUT(zuord.SIFA, komm.Unbestimmt, aktion.Down, tastatur_schalterposition=0),
        165: messages.INPUT(zuord.SIFA, komm.Unbestimmt, aktion.Up, tastatur_schalterposition=0),
        170: messages.INPUT(indusi_stoerschalter=messages.SCHALTER.AUS),
        180: messages.INPUT(lzb_stoerschalter=messages.SCHALTER.AUS),
        210: messages.INPUT(lzb_stoerschalter=messages.SCHALTER.EIN),
        220: messages.INPUT(indusi_stoerschalter=messages.SCHALTER.EIN),
    }

    known_states = {}
    wait_cur = 0
    while True:
        # handle incremental updates from Zusi to print only changed state to console
        if client.local_state_changed.is_set():
            client.local_state_changed.clear()
            for message, state in client.local_state.items():
                current_state = state._asdict()
                if message not in known_states:
                    known_states[message] = current_state
                    log.info("%s state: %s" % (message.__name__, suppress_none_values(current_state)))
                    continue
                try:
                    updates = sorted(set(current_state.items()) - set(known_states[message].items()))
                except TypeError:
                    # Happens when submessages have lists, assume all changed if one part changed
                    if known_states[message] != current_state:
                        updates = current_state
                    else:
                        continue
                if updates:
                    known_states[message] = current_state
                    log.info("%s changes: %s" % (message.__name__, updates))
        await asyncio.sleep(0.1)
        wait_cur += 1
        if wait_cur in simulated_interactions:
            msg = simulated_interactions[wait_cur]
            log.info("Sending new action %s: %s" % (msg.__class__.__name__, str(suppress_none_values(msg._asdict()))))
            client.send_input(msg)
        if wait_cur > WAIT_MAX:
            log.info("Time is up, leaving")
            break

tasks = []
async def main(ip, port):
    client = ZusiClient(ip, port, "pyzusi3 demo", "1.0")
    client.request_status(displays=[
            messages.FAHRPULT_ANZEIGEN.GESCHWINDIGKEIT,
            messages.FAHRPULT_ANZEIGEN.STATUS_SIFA,
            messages.FAHRPULT_ANZEIGEN.STATUS_ZUGBEEINFLUSSUNG
        ],
        control=True,
        programdata=[
            messages.PROGRAMMDATEN.ZUGDATEI,
            messages.PROGRAMMDATEN.ZUGNUMMER,
            messages.PROGRAMMDATEN.LADEPAUSE,
        ]
    )

    main_task = asyncio.create_task(client.connect())
    tasks.append(main_task)

    interact_task = asyncio.create_task(zusi_user_interact(client))
    tasks.append(interact_task)

    while True:
        all_tasks_done = True
        for task in tasks:
            if task.done():
                exc = task.exception()
                if exc:
                    log.exception(exc, exc_info=exc)
                    pass
            else:
                all_tasks_done = False
        if all_tasks_done:
            break
        await asyncio.sleep(1)

if __name__ == "__main__":
    run_loop = asyncio.new_event_loop()
    run_loop.set_exception_handler(handle_exception)
    run_loop.create_task(main(ZUSI_IP, ZUSI_PORT))
    run_loop.run_forever()



import unittest

from pyzusi3.nodes import BasicNode, ContentType, StreamDecoder

class TestManualExample1(unittest.TestCase):
    def setUp(self):
        self.bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x02\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x02\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x46\x61\x68\x72\x70\x75\x6C\x74' + \
            b'\x05\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x32\x2E\x30' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'
        self.command = BasicNode(id=1) # Verbindungsaufbau
        hellonode = BasicNode(id=1) # HELLO-Befehl
        self.command.children.append(hellonode)
        hellonode.children.append(BasicNode(id=1, content=2, contenttype=ContentType.WORD)) # Protokollversion
        hellonode.children.append(BasicNode(id=2, content=2, contenttype=ContentType.WORD)) # Clienttyp Fahrpult
        hellonode.children.append(BasicNode(id=3, content="Fahrpult", contenttype=ContentType.STRING)) # Clientname
        hellonode.children.append(BasicNode(id=4, content="2.0", contenttype=ContentType.STRING)) # Clientversion

    def test_encoding(self):
        self.assertEqual(self.command.encode(), self.bytes_written)

    def test_msg_decoding(self):
        decoder = StreamDecoder()
        result = b''
        for decoded_tree in decoder.decode(self.bytes_written):
            result += decoded_tree.encode()
        self.assertEqual(self.bytes_written, result)

class TestManualExample2(unittest.TestCase):
    def setUp(self):
        self.bytes_written = b'' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x09\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x33\x2E\x30\x2E\x31\x2E\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x30' + \
            b'\x03\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\x00' + \
            b'\x0A\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\x00\x00\x00\x00\xD0\x35\xE4\x40' + \
            b'\xFF\xFF\xFF\xFF' + \
            b'\xFF\xFF\xFF\xFF'

        self.command = BasicNode(id=1) # Verbindungsaufbau
        hellonode = BasicNode(id=2) # ACK_HELLO-Befehl
        self.command.children.append(hellonode)
        hellonode.children.append(BasicNode(id=1, content="3.0.1.0", contenttype=ContentType.STRING)) # Zusi Version
        hellonode.children.append(BasicNode(id=2, content="0", contenttype=ContentType.STRING)) # Verbindungsinfo
        hellonode.children.append(BasicNode(id=3, content=0, contenttype=ContentType.SHORTINT)) # Ergebnis: 0=OK
        hellonode.children.append(BasicNode(id=4, content=41390.5, contenttype=ContentType.DOUBLE)) # Fahrplananfangszeit

    def test_encoding(self):
        self.assertEqual(self.command.encode(), self.bytes_written)

    def test_msg_decoding(self):
        decoder = StreamDecoder()
        result = b''
        for decoded_tree in decoder.decode(self.bytes_written):
            result += decoded_tree.encode()
        self.assertEqual(self.bytes_written, result)


class TestRealWorldStreamdata(unittest.TestCase):
    def test_zusilm_message(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xa9\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x08\x00\x00\x00' + \
            b'\x01\x00' + \
            b'ZD_MTD' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x02' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\xe8\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'\xa9\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xa9\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\n\x00\x00\x00' + \
            b'\x01\x00' + \
            b'ZD_EBuLa' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x02' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\xaa\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'v\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xa9\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\x01\x00' + \
            b'\x0c\x00\x00\x00' + \
            b'\x01\x00' + \
            b'ZD_Zugfunk' + \
            b'\x03\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x02' + \
            b'\x04\x00\x00\x00' + \
            b'\x03\x00' + \
            b'\xc9\x00' + \
            b'\x04\x00\x00\x00' + \
            b'\x04\x00' + \
            b'h\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'
        decoder = StreamDecoder()
        result = b''
        for decoded_tree in decoder.decode(bytes_written):
            result += decoded_tree.encode()
        self.assertEqual(bytes_written, result)

    def test_zusilm2_msg(self):
        bytes_written = b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff' + \
            b'\x00\x00\x00\x00' + \
            b'\x02\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\n\x00' + \
            b'\x06\x00\x00\x00' + \
            b'\xa9\x00' + \
            b'\x00\x00\x00\x00' + \
            b'\xff\xff\xff\xff' + \
            b'\xff\xff\xff\xff'
        decoder = StreamDecoder()
        result = b''
        for decoded_tree in decoder.decode(bytes_written):
            result += decoded_tree.encode()
        self.assertEqual(bytes_written, result)



import asyncio
import logging
import threading

from pyzusi3.nodes import AsyncStreamDecoder
from pyzusi3 import messages
from pyzusi3.messagecoders import MessageDecoder, encode_obj

reader_log = logging.getLogger("pyzusi3.client.reader")
writer_log = logging.getLogger("pyzusi3.client.writer")
updater_log = logging.getLogger("pyzusi3.client.updater")
client_log = logging.getLogger("pyzusi3.client")


# Debug function
LOG_MSG_UPDATES = False


def suppress_none_values(dictionary):
    """Helper to just return items from dictionary where the value is not None"""

    cleaned_collection = {}
    for key, value in dictionary.items():
        if value is not None:
            cleaned_collection[key] = value

    return cleaned_collection


class ZusiClient:
    def __init__(self, ip, port, clientname="pyZusi3 Fahrpult", clientversion="1.0"):
        self.ip = ip
        self.port = port
        self.clientname = clientname
        self.clientversion = clientversion

        self.task_registry = {}
        
        self.send_messagequeue = asyncio.Queue()
        self.receive_messagequeue = asyncio.Queue()
        self.send_messagequeue_lock = threading.Lock()

        self.local_state = {}
        self.message_update_events = {}
        self.local_state_changed = asyncio.Event()

        self.connected = False
        self.requested_status = None

    async def _decode_bytes(self, stream_bytes):
        """Generator to get decoded messages from a stream"""

        decoder = AsyncStreamDecoder()
        decoded_tree = decoder.decode(stream_bytes)
        nodes = []
        async for node in decoded_tree:
            messagedecoder = MessageDecoder()
            yield messagedecoder.parse(node)      

    async def _zusi_writer(self, write_stream):
        """Queue worker to send messages to Zusi on the stream"""

        while True:
            msg = await self.send_messagequeue.get()
            writer_log.debug("Sending msg to Zusi: %s %s" % (msg.__class__.__name__, str(suppress_none_values(msg._asdict()))))
            write_stream.write(encode_obj(msg).encode())

    async def _zusi_reader(self, reader):
        """Queue worker to process received messages from Zusi on given stream"""

        msg_reader = self._decode_bytes(reader)

        while True:
            basemessage, submessages = await msg_reader.__anext__()
            if basemessage is not None:
                reader_log.debug("Got basemessage from Zusi: %s %s" % (basemessage.__class__.__name__, str(suppress_none_values(basemessage._asdict()))))
                self.receive_messagequeue.put_nowait(basemessage)
            for message in submessages:
                reader_log.debug("Got submessage from Zusi: %s %s" % (message.__class__.__name__, str(suppress_none_values(message._asdict()))))
                self.receive_messagequeue.put_nowait(message)

    async def _update_local_state(self):
        updater_log.info("Awaiting new states")
        while True:
            msg = await self.receive_messagequeue.get()
            updater_log.debug("Got new data for %s" % str(type(msg)))
            if msg is None:
                continue

            updater_log.debug("Checking for known state")
            msgtype = type(msg)
            if msgtype not in self.local_state:
                updater_log.debug("Just saving whole state, was unknown")
                self.local_state[msgtype] = msg
                self.message_update_events[msgtype] = asyncio.Event()
                self.local_state_changed.set()
                continue

            updater_log.debug("Checking for state increment updates")
            updated_keys = suppress_none_values(msg._asdict())
            if not updated_keys:
                updater_log.debug("None found, skipping")
                continue
            known_msg_state = self.local_state[msgtype]
            unchanged_keys = []
            for key in updated_keys.keys():
                if getattr(known_msg_state, key) == updated_keys[key]:
                    unchanged_keys.append(key)
            for key in unchanged_keys:
                del updated_keys[key]
            if not updated_keys:
                updater_log.debug("Just same content, skipping")
                continue

            updater_log.debug("Updated keys: %s" % updated_keys)
            self.local_state[msgtype] = known_msg_state._replace(**updated_keys)
            self.message_update_events[msgtype].set()
            self.local_state_changed.set()

            if not LOG_MSG_UPDATES:
                continue

            # normal messages.FAHRPULT_ANZEIGEN states
            data = str(suppress_none_values(self.local_state[msgtype]._asdict()))
            if isinstance(msg, messages.DATA_FTD):
                updater_log.warning("Updated general train data: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_NOTBREMSSYSTEM
            elif isinstance(msg, messages.STATUS_NOTBREMSSYSTEM):
                updater_log.warning("Updated state for emer brakes: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_SIFA
            elif isinstance(msg, messages.STATUS_SIFA):
                updater_log.warning("Updated state for Sifa: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_ZUGBEEINFLUSSUNG
            elif isinstance(msg, messages.STATUS_ZUGBEEINFLUSSUNG_GRUND):
                updater_log.warning("Updated state for ZB basic: %s" % data)
            elif isinstance(msg, messages.STATUS_INDUSI_EINSTELLUNGEN):
                updater_log.warning("Updated state for ZB settings: %s" % data)
            elif isinstance(msg, messages.STATUS_INDUSI_BETRIEBSDATEN):
                updater_log.warning("Updated state for ZB state: %s" % data)
            elif isinstance(msg, messages.STATUS_ETCS_EINSTELLUNGEN):
                updater_log.warning("Updated state for ETCS settings: %s" % data)
            elif isinstance(msg, messages.STATUS_ETCS_BETRIEBSDATEN):
                updater_log.warning("Updated state for ETCS state: %s" % data)
            elif isinstance(msg, messages.STATUS_ZUB_EINSTELLUNGEN):
                updater_log.warning("Updated state for ZUB settings: %s" % data)
            elif isinstance(msg, messages.STATUS_ZUB_BETRIEBSDATEN):
                updater_log.warning("Updated state for ZUB state: %s" % data)
            elif isinstance(msg, messages.STATUS_ZBS_EINSTELLUNGEN):
                updater_log.warning("Updated state for ZBS settings: %s" % data)
            elif isinstance(msg, messages.STATUS_ZBS_BETRIEBSDATEN):
                updater_log.warning("Updated state for ZBS state: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_TUEREN
            elif isinstance(msg, messages.STATUS_TUEREN):
                updater_log.warning("Updated state for doors: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_FAHRZEUG
            elif isinstance(msg, messages.STATUS_FAHRZEUG):
                updater_log.warning("Updated state for vehicle: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_ZUGVERBAND
            elif isinstance(msg, messages.STATUS_ZUGVERBAND):
                updater_log.warning("Updated state for train: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_WEICHEN
            elif isinstance(msg, messages.STATUS_WEICHEN):
                updater_log.warning("Updated state for switch: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_LM_ZUSIPISPLAY
            elif isinstance(msg, messages.STATUS_LM_ZUSIDISPLAY):
                updater_log.warning("Updated state for zusidisplay buttons: %s" % data)
            # messages.FAHRPULT_ANZEIGEN.STATUS_ZUGFAHRDATEN
            elif isinstance(msg, messages.STATUS_ZUGFAHRDATEN):
                updater_log.warning("Updated state for train wagon data: %s" % data)
            elif isinstance(msg, messages.DATA_PROG):
                updater_log.warning("Updated simulation data: %s" % data)
            # messages.bedienung = True
            elif isinstance(msg, messages.DATA_OPERATION):
                updater_log.warning("Updated input data: %s" % data)
            else:
                updater_log.error("Unsorted message: type: %s content: %s" % (type(msg), data))

    def request_status(self, displays=[], control=False, programdata=[]):
        self.requested_status = messages.NEEDED_DATA(
            anzeigen=displays,
            bedienung=control,
            programmdaten=programdata
        )
        if self.connected and 'updater' in self.task_registry:
            self.send_messagequeue.put_nowait(self.requested_status)

    def send_input(self, input):
        if not isinstance(input, messages.INPUT):
            raise ValueError("Input needs to be in message INPUT format")
        
        if not self.connected:
            raise ConnectionError("Not connected to Zusi")
        with self.send_messagequeue_lock:
            self.send_messagequeue.put_nowait(input)

    async def connect(self):
        client_log.info("Connecting to Zusi3")
        tcpreader, tcpwriter = await asyncio.open_connection(self.ip, self.port)
        self.connected = True

        client_log.info("Starting reader and writer task")
        reader_task = asyncio.create_task(self._zusi_reader(tcpreader))
        self.task_registry['reader'] = reader_task
        writer_task = asyncio.create_task(self._zusi_writer(tcpwriter))
        self.task_registry['writer'] = writer_task

        client_log.info("Sending HELLO message")
        msg = messages.HELLO(2, messages.ClientTyp.FAHRPULT, self.clientname, self.clientversion)
        self.send_messagequeue.put_nowait(msg)

        client_log.info("Waiting for response")
        msg = await self.receive_messagequeue.get()
        if not (isinstance(msg, messages.ACK_HELLO) and msg.status == 0):
            client_log.error("Zusi did not report success for HELLO")
            return

        client_log.info("Waiting for user status to request content")
        while self.requested_status is None:
            await asyncio.sleep(0.1)
        self.send_messagequeue.put_nowait(self.requested_status)

        client_log.info("Awaiting status response ok from Zusi")
        msg = await self.receive_messagequeue.get()
        if not (isinstance(msg, messages.ACK_NEEDED_DATA) and msg.status == 0):
            client_log.error("Zusi did not report success for NEEDED_DATA")
            return

        client_log.info("Creating task to update local state when Zusi reports data")
        update_task = asyncio.create_task(self._update_local_state())
        self.task_registry['updater'] = update_task

        # wait for communication to finish
        await asyncio.wait([reader_task])

        # cancel all other tasks and close connection
        writer_task.cancel()
        update_task.cancel()

        tcpwriter.close()
        await tcpwriter.wait_closed()

        self.connected = False

        client_log.info("Connection closed")


import asyncio
import logging
import math
import time
import threading

from PySide6.QtWidgets import QMainWindow
from PySide6.QtCore import QTimer, Slot
from pyzusi3.client import ZusiClient
from pyzusi3 import messages as zusimsg

from ui_form import Ui_MainWindow

class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.zusiClient = None
        self.zusiTask = None
        self.zusiLoop = asyncio.new_event_loop()
        self.zusiThread = None

        self.led_init = True
        self.reset_led()
        self.reset_textfields()

        self.timer_ui_update = QTimer()
        self.timer_ui_update.timeout.connect(self.update_ui)

    def reset_led(self):
        self.previous_led_state = {
            'u': None,
            'm': None,
            'o': None,
            '40': None,
            '500': None,
            '1000': None,
            'sifa': None
        }
        self.update_leds()

    def get_leds_state(self):
        if self.zusiClient is None:
            return {
                'u': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                'm': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                'o': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                '40': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                '500': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                '1000': zusimsg.LMZUSTAND_MIT_INVERS.AUS,
                'sifa': zusimsg.LMZUSTAND_MIT_INVERS.AUS
            }

        if zusimsg.STATUS_INDUSI_BETRIEBSDATEN not in self.zusiClient.local_state:
            return self.previous_led_state

        indusi_state = self.zusiClient.local_state[zusimsg.STATUS_INDUSI_BETRIEBSDATEN]

        led_state = {
            'u': indusi_state.lm_u,
            'm': indusi_state.lm_m,
            'o': indusi_state.lm_o,
            '40': indusi_state.lm_befehl,
            '500': indusi_state.lm_500hz,
            '1000': indusi_state.lm_1000hz
        }

        if zusimsg.STATUS_SIFA not in self.zusiClient.local_state:
            sifa_state = zusimsg.LMZUSTAND.AUS
        else:
            sifa_state = self.zusiClient.local_state[zusimsg.STATUS_SIFA].lm
        led_state['sifa'] = sifa_state

        return led_state

    def update_ui(self):
        self.setUpdatesEnabled(False)
        self.update_leds()
        self.update_text()
        self.setUpdatesEnabled(True)

    def update_leds(self):
        if not self.led_init:
            if self.zusiClient is None:
                return
        else:
            self.led_init = False

        current_time = time.time()
        # create 500ms blink
        blink_on = (current_time - int(current_time) >= 0.5)
        new_state = self.get_leds_state()
        current_state = self.previous_led_state
        led_off_states = [
            None,
            zusimsg.LMZUSTAND.AUS,
            zusimsg.LMZUSTAND_MIT_INVERS.AUS,
        ]
        if blink_on:
            led_off_states.append(zusimsg.LMZUSTAND_MIT_INVERS.BLINKEND_INVERS)
        else:
            led_off_states.append(zusimsg.LMZUSTAND.BLINKEN)
            led_off_states.append(zusimsg.LMZUSTAND_MIT_INVERS.BLINKEND)

        for statevar, uielement, on_color, off_color in [
            ('u', self.ui.lm_u, 'dodgerblue', 'mediumblue'),
            ('m', self.ui.lm_m, 'dodgerblue', 'mediumblue'),
            ('o', self.ui.lm_o, 'dodgerblue', 'mediumblue'),
            ('40', self.ui.lm_40, 'white', 'gray'),
            ('500', self.ui.lm_500, 'orangered', 'maroon'),
            ('1000', self.ui.lm_1000, 'gold', 'rgb(136, 73, 16)'),
            ('sifa', self.ui.lm_sifa, 'white', 'gray')
        ]:
            if current_state[statevar] != new_state[statevar] or \
                new_state[statevar] in [zusimsg.LMZUSTAND.BLINKEN, zusimsg.LMZUSTAND_MIT_INVERS.BLINKEND_INVERS]:

                lamp_state = new_state[statevar]
                if lamp_state in led_off_states:
                    uielement.setStyleSheet("color: black; background-color: %s;" % off_color)
                else:
                    uielement.setStyleSheet("color: black; background-color: %s;" % on_color)

    def update_text(self):
        if self.zusiClient is None:
            return

        if zusimsg.DATA_PROG in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.DATA_PROG]
    
            self.ui.fahrplan.setText(str(state.zugdateiname))
            self.ui.zugnummer.setText(str(state.zugnummer))
            self.ui.ladezustand.setText(str(state.ladepause))
        
        if zusimsg.DATA_FTD in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.DATA_FTD]

            if state.uhrzeit_stunde is not None and state.uhrzeit_minute is not None:
                stunde = int(state.uhrzeit_stunde)
                minute = state.uhrzeit_minute
                sekunde = int((minute - int(minute))*0.6*100)
                minute = int(minute)
                self.ui.uhrzeit.setText("%s:%s:%s" % (str(stunde).zfill(2), str(minute).zfill(2), str(sekunde).zfill(2)))

            self.ui.geschwindigkeit.setText(str(int(round(state.geschwindigkeit*3.6))))
            if self.ui.sollgeschwindigkeit_check.isChecked():
                self.ui.sollgeschwindigkeit.setText(str(int(state.streckenvmax*3.6)))
            else:
                self.ui.sollgeschwindigkeit.clear()

        if zusimsg.STATUS_ZUGFAHRDATEN in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_ZUGFAHRDATEN]

            if state.fahrzeuge:
                fahrzeug = state.fahrzeuge[0]
                self.ui.druckhll.setText(str(round(fahrzeug.hll_druck, 1)))
                self.ui.bremsdruck.setText(str(round(fahrzeug.bremszylinderdruck, 1)))

        if zusimsg.STATUS_ZUGBEEINFLUSSUNG_GRUND in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_ZUGBEEINFLUSSUNG_GRUND]

            self.ui.indusibauart.setText(state.bauart)

        if zusimsg.STATUS_INDUSI_EINSTELLUNGEN in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_INDUSI_EINSTELLUNGEN]

 
        if zusimsg.STATUS_INDUSI_BETRIEBSDATEN in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_INDUSI_BETRIEBSDATEN]

            self.ui.indusistatus.setText(str(state.zustand.name if state.zustand is not None else ''))
            self.ui.indusizwangsbremsung.setText(str(state.zwangsbremsung.name if state.zwangsbremsung is not None else ''))
            self.ui.indusizusatzinfo.setText(str(state.zusatzinfo_melderbild.name if state.zusatzinfo_melderbild is not None else ''))

        if zusimsg.STATUS_SIFA in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_SIFA]

            self.ui.sifabauart.setText(state.bauart)
            status = "NORMAL"
            if state.lm == zusimsg.LMZUSTAND.AN:
                status = "LAMPE AN"
            if state.hupe is not None and state.hupe != zusimsg.STATUS_SIFA_HUPE.HUPE_AUS:
                status = state.hupe.name
            self.ui.sifastatus.setText(status)

        if zusimsg.STATUS_NOTBREMSSYSTEM in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_NOTBREMSSYSTEM]

            self.ui.nbuebauart.setText(state.bauart)
            self.ui.nbuestatus.setText(str(state.status.name if state.status is not None else ''))

        if zusimsg.STATUS_TUEREN in self.zusiClient.local_state:
            state = self.zusiClient.local_state[zusimsg.STATUS_TUEREN]

            self.ui.tuerenlinks.setText(str(state.links.name if state.links is not None else ''))
            self.ui.tuerenrechts.setText(str(state.rechts.name if state.rechts is not None else ''))

    def reset_textfields(self):
        self.ui.geschwindigkeit.clear()
        self.ui.druckhll.clear()
        self.ui.indusistatus.clear()
        self.ui.sifastatus.clear()
        self.ui.uhrzeit.clear()
        self.ui.nbuestatus.clear()
        self.ui.sifabauart.clear()
        self.ui.indusibauart.clear()
        self.ui.nbuebauart.clear()
        self.ui.indusizwangsbremsung.clear()
        self.ui.indusizusatzinfo.clear()
        self.ui.sollgeschwindigkeit.clear()
        self.ui.tuerenlinks.clear()
        self.ui.tuerenrechts.clear()
        self.ui.zugnummer.clear()
        self.ui.fahrplan.clear()
        self.ui.ladezustand.clear()

    @Slot()
    def on_connectButton_clicked(self):
        self.ui.connectButton.setEnabled(False)
        self.timer_ui_update.start(500)
        self.zusiThread = threading.Thread(target=lambda: self.run_zusi_loop(), daemon=True)
        self.zusiThread.start()

    def run_zusi_loop(self):
        asyncio.set_event_loop(self.zusiLoop)
        self.zusiLoop.call_soon(lambda: self.connect_zusi())
        self.zusiLoop.run_forever()

    def connect_zusi(self):
        self.zusiClient = ZusiClient(self.ui.zusi_ip.text(), 1436, "ZusiData Display", "1.0")
        self.zusiClient.request_status(
            displays=[
                zusimsg.FAHRPULT_ANZEIGEN.GESCHWINDIGKEIT,
                zusimsg.FAHRPULT_ANZEIGEN.UHRZEIT_STUNDE,
                zusimsg.FAHRPULT_ANZEIGEN.UHRZEIT_MINUTE,
                zusimsg.FAHRPULT_ANZEIGEN.STRECKENVMAX,

                zusimsg.FAHRPULT_ANZEIGEN.STATUS_ZUGFAHRDATEN,
                zusimsg.FAHRPULT_ANZEIGEN.STATUS_NOTBREMSSYSTEM,
                zusimsg.FAHRPULT_ANZEIGEN.STATUS_TUEREN,
                zusimsg.FAHRPULT_ANZEIGEN.STATUS_SIFA,
                zusimsg.FAHRPULT_ANZEIGEN.STATUS_ZUGBEEINFLUSSUNG
            ],
            programdata=[
                zusimsg.PROGRAMMDATEN.ZUGDATEI,
                zusimsg.PROGRAMMDATEN.ZUGNUMMER,
                zusimsg.PROGRAMMDATEN.LADEPAUSE,
            ]
        )
        loop = asyncio.get_event_loop()
        loop.create_task(self.zusiClient.connect())



from collections import defaultdict
import logging
from enum import Enum
from operator import attrgetter
import struct
from pyzusi3.exceptions import MissingLowLevelParameterError

from pyzusi3.messages import ParameterId, message_index, lowlevel_parameters, ContentType
from pyzusi3.nodes import BasicNode

def print_nodetree(node, level=0):
    print("%s%s:%s" % (" " * level, node.id, node.content))
    for child in node.children:
        print_nodetree(child, level + 1)


def decode_data(data, contenttype, enumtype):
    if data is None:
        return

    if contenttype == ContentType.BYTE:
        result = int.from_bytes(data, byteorder='little')
    elif contenttype == ContentType.SHORTINT:
        result = struct.unpack("<b", data)[0]
    elif contenttype == ContentType.WORD:
        result = struct.unpack("<H", data)[0]
    elif contenttype == ContentType.SMALLINT:
        result = struct.unpack("<h", data)[0]
    elif contenttype == ContentType.INTEGER:
        result = struct.unpack("<i", data)[0]
    elif contenttype == ContentType.CARDINAL:
        result = struct.unpack("<I", data)[0]
    elif contenttype == ContentType.INTEGER64BIT:
        result = struct.unpack("<q", data)[0]
    elif contenttype == ContentType.SINGLE:
        result = struct.unpack("<f", data)[0]
    elif contenttype == ContentType.DOUBLE:
        result = struct.unpack("<d", data)[0]
    elif contenttype == ContentType.STRING:
        result = data.decode("latin1")
    elif contenttype == ContentType.FILE or contenttype == ContentType.RAW:
        result = data
    else:
        raise NotImplementedError("Unknown content type %s given" % contenttype)

    if enumtype is not None:
        result = enumtype(result)

    return result


class MessageDecoder:
    def __init__(self) -> None:
        self.log = logging.getLogger("pyzusi3.messagecoders.MessageDecoder")
        self.reset()

    def reset(self):
        self.message_class = None
        self.submessages = []
        self.mapped_parameters = {}
        self.lowlevel_parameter = None
    
    def parse(self, root_node, start_level=1):
        if not self.message_class:
            self.init_msg_params(root_node)

        self.map_parameters(root_node, self.message_pid, start_level)

        return self.finalize()

    def init_msg_params(self, root_node):
        self.message_class, self.message_pid = self.find_messageclass(root_node)
        self.lowlevel_parameter = lowlevel_parameters[self.message_class]

    def finalize(self):
        basemessage = self.message_class(**self.mapped_parameters)
        return basemessage, self.submessages

    def find_messageclass(self, root_node):
        current_pid = ParameterId()
        current_node = root_node
        current_level = 1

        while current_level < 7:
            params = {'id' + str(current_level): current_node.id}
            current_pid = current_pid._replace(**params)
            if current_pid in message_index:
                break
            current_level += 1
            if len(current_node.children) == 0:
                # bug detector, should never happen
                break
            current_node = current_node.children[0]

        if current_pid not in message_index:
            raise NotImplementedError("Message for index %s not yet implemented" % str(current_pid))
        
        return message_index[current_pid], current_pid

    def map_parameters(self, current_node, current_pid, current_level):
        if current_pid == ParameterId(2, 10, 169) and current_node.content:
            # whitelist bug in FTD message
            # https://forum.zusi.de/viewtopic.php?f=55&t=18246
            return

        mapping_parameter = sorted([param for param in self.lowlevel_parameter if param.parameterid == current_pid])
        if len(mapping_parameter) > 1:
            raise NotImplementedError("Parameter %s is not unique for %s, programming error!" % (current_pid, self.message_class))
        if not len(mapping_parameter):
            self.log.debug("Parameter %s is not known for %s, checking for submessage" % (current_pid, self.message_class))
            check_param_index = None
            submessage_class = None
            for i in range(len(current_pid), 1, -1):
                check_param_index = ParameterId(*current_pid[:i])
                if check_param_index in message_index:
                    submessage_class = message_index[check_param_index]
                    if submessage_class == self.message_class:
                        submessage_class = None
                        break
                    self.log.debug("Found submessage of type %s" % submessage_class)
                    break
            if submessage_class is None:
                self.log.warning("Parameter %s is not known for %s and no submessage, discarding" % (current_pid, self.message_class))
                return

            # Handle submessage as we found a new root id for it
            submsg_decoder = MessageDecoder()
            submsg_decoder.message_class = submessage_class
            submsg_decoder.message_pid = check_param_index
            submsg_decoder.lowlevel_parameter = lowlevel_parameters[submessage_class]
            basemsg, submsgs = submsg_decoder.parse(current_node, current_level)
            self.submessages.append(basemsg)
            if submsgs:
                self.submessages.extend(submsgs)
            return

        mapping_parameter = mapping_parameter[0]
        if mapping_parameter.contenttype is BasicNode and mapping_parameter.multipletimes is not None:
            # handle multipletimes-style submessages
            multi_msg_class = mapping_parameter.multipletimes
            multi_msg_decoder = MessageDecoder()
            multi_msg_decoder.message_class = multi_msg_class
            multi_msg_decoder.message_pid = current_pid
            multi_msg_decoder.lowlevel_parameter = lowlevel_parameters[multi_msg_class]

            for child_node in current_node.children:
                sub_pid = {'id' + str(current_level + 1): child_node.id}          
                multi_msg_decoder.map_parameters(child_node, current_pid._replace(**sub_pid), current_level + 1)
            basemsg, submsgs = multi_msg_decoder.finalize()

            paramname = mapping_parameter.parametername
            if paramname not in self.mapped_parameters:
                self.mapped_parameters[paramname] = []
            self.mapped_parameters[paramname].append(basemsg)

            return
        if current_node.content:
            # Decode binary content
            decoded_content = decode_data(current_node.content, mapping_parameter.contenttype, mapping_parameter.enumtype)
            if mapping_parameter.multipletimes == True:
                if mapping_parameter.parametername not in self.mapped_parameters:
                    self.mapped_parameters[mapping_parameter.parametername] = []
                self.mapped_parameters[mapping_parameter.parametername].append(decoded_content)
            else:
                self.mapped_parameters[mapping_parameter.parametername] = decoded_content
        elif current_node.nodeasbool:
            # Tree just has an empty node but acts as True value
            if mapping_parameter.nodeasbool == True:
                self.mapped_parameters[mapping_parameter.parametername] = True
        for child_node in current_node.children:
            params = {'id' + str(current_level + 1): child_node.id}
            child_pid = current_pid._replace(**params)
            self.map_parameters(child_node, child_pid, current_level + 1)


def level_for_parameterid(parameterid):
    for i in range(6, 0, -1):
        if getattr(parameterid, "id" + str(i), None) is not None:
            return i
    return 0

def strip_level_from_parameterid(level, parameterid):
    new_paramid = parameterid
    for i in range(6, level-1, -1):
        nulled_id = {'id' + str(i): None}
        new_paramid = new_paramid._replace(**nulled_id)
    return new_paramid

def is_subparameter(parentparameterid, subparameterid):
    parent_level = level_for_parameterid(parentparameterid)
    stripped_subparameterid = strip_level_from_parameterid(parent_level + 1, subparameterid)
    return parentparameterid == stripped_subparameterid

def encode_obj(obj):
    if type(obj) not in lowlevel_parameters:
        raise MissingLowLevelParameterError("No known %s in low level parameter encoding list" % type(obj))

    parametertree = lowlevel_parameters[type(obj)]
    if parametertree is not None:
        parametertree = sorted(parametertree, key=attrgetter('parameterid'))

    root_node = None
    current_node = None
    current_level = 1
    seen_treenodes = {}
    for parameter in parametertree:
        # find parent node to attach to
        parent_node = None
        for i in range(level_for_parameterid(parameter.parameterid), 0, -1):
            parent_paramid = strip_level_from_parameterid(i, parameter.parameterid)
            if parent_paramid in seen_treenodes:
                parent_node = seen_treenodes[parent_paramid]
                current_level = i
                break

        if parameter.contenttype is BasicNode:
            if parameter.multipletimes is None or \
                (parameter.multipletimes is not None and parameter.multipletimes == type(obj)):
                if parameter.nodeasbool and getattr(obj, parameter.parametername, False) != True:
                    continue
                current_node = BasicNode(id=getattr(parameter.parameterid, 'id' + str(current_level)), parent_node=parent_node, nodeasbool=parameter.nodeasbool)
                if current_node.parent_node is None:
                    root_node = current_node
                else:
                    current_node.parent_node.children.append(current_node)
                seen_treenodes[parameter.parameterid] = current_node
                continue

            parent_tree_pid = parameter.parameterid
            for item in getattr(obj, parameter.parametername, []):
                encoded_item = encode_obj(item)
                def locate_subtree(tree, built_parameterid):
                    tree_level = level_for_parameterid(built_parameterid)
                    new_pidindex = {'id' + str(tree_level + 1): tree.id}
                    new_parameterid = built_parameterid._replace(**new_pidindex)
                    if new_parameterid == parent_tree_pid:
                        return tree
                    found_subtree = None
                    for child_tree in tree.children:
                        found_subtree = found_subtree or locate_subtree(child_tree, new_parameterid)
                        if found_subtree is not None:
                            break
                    return found_subtree

                subtree = locate_subtree(encoded_item, ParameterId())
                if subtree is None:
                    raise ValueError("Tree for parameter %s inconsistent, did not find common root" % str(item))
                subtree.parent_node = current_node
                current_node.children.append(subtree)
            continue

        parameter_value = getattr(obj, parameter.parametername, None)
        if parameter_value is None:
            continue
        node_id = getattr(parameter.parameterid, "id" + str(current_level))
        node_content = parameter_value
        if type(node_content) == bytes and parameter.contenttype not in [ContentType.FILE, ContentType.RAW]:
            node_contenttype = ContentType.RAW
        else:
            node_contenttype = parameter.contenttype
        if isinstance(node_content, list) and parameter.contenttype is not BasicNode and parameter.multipletimes is True:
           for entry in node_content:
               if isinstance(entry, Enum):
                   entry_content = entry.value
               else:
                   entry_content = entry
               new_node = BasicNode(id=node_id, content=entry_content, contenttype=node_contenttype, parent_node=current_node)
               if new_node.content is not None:
                   current_node.children.append(new_node)
        else:
            if isinstance(node_content, Enum):
                node_content = node_content.value
            selected_parentnode = seen_treenodes[parent_paramid]                
            new_node = BasicNode(id=node_id, content=node_content, contenttype=node_contenttype, parent_node=selected_parentnode)
            if new_node.content is not None:
                selected_parentnode.children.append(new_node)

    def optimize_tree(node):
        """Removes all empty child nodes"""

        tree_has_changed = False

        new_children = []
        for child in node.children:
            if child.children:
                tree_has_changed = optimize_tree(child) or tree_has_changed
            if child.children or child.content is not None or child.nodeasbool:
               new_children.append(child)

        tree_has_changed = tree_has_changed or len(node.children) != len(new_children)
        node.children = new_children
    
        return tree_has_changed

    changed = True
    while changed:
        changed = optimize_tree(root_node)

    return root_node
